<!--
  This file contains the templates specific to generating the c wrapper code.
  -->

<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">

  <xsl:import href="common.xsl"/>
  <xsl:output method="text"/>



 <!--
    Root level.  This gets things going.
    -->
  <xsl:template match="/">
    <xsl:apply-templates select="library"/>
  </xsl:template>



  <!--
    Library level (can there ever be more than one?).
    -->
  <xsl:template match="library">
    <xsl:call-template name="open_file"/>
    <xsl:call-template name="generate_c_incs"/>
    <xsl:call-template name="extern_c_pre"/>

    <xsl:apply-templates select="scope/descendant::method" mode="decl_fc_c"/>

    <xsl:call-template name="extern_c_post"/>
    <xsl:call-template name="close_file"/>
  </xsl:template>



  <!--
    This template generates the name of a wrapper function.  Names can be
    changed in several ways.  This template does the following:
      o prepends the class name to any method internally.  the decision
        of whether to prepend the class name to the fortran function name
        is set in the user preferences file (except ctors/dtors, they
        always have the class name prepended).
      o does a lookup to see if there is a specific abbreviation that
        should be used from the user preferences file for the class or
        method name.
      o appends the position number of the method to the end.
      o does any dictionary renaming that's appropriate.
      o checks the length of the resulting name to make sure it's legal.
    
    This template is very similar to the one used for fortran func names.    
    -->
  <xsl:template name="gen_func_name">

    <!-- since we need to check the resulting length of the generated name,
         we do all the output of this template into a variable, which we can
         just then check at the end -->
    <xsl:variable name="final_name">

      <!-- find out what the original function name is -->
      <xsl:variable name="func_name" select="@name"/>

      <!-- if this is a class method, then we need to look up the alias for
           the class from the user preference xml file and then prefix the function
           name with it -->
      <xsl:if test="ancestor::structure">

        <!-- find out what the original class name is -->
        <xsl:variable name="class_name" select="ancestor::structure/@name"/>

        <!-- lookup the aliases for the class/method names -->
        <xsl:variable name="class_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/@alias"/>
        <xsl:variable name="class_id" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/@mangle"/>
        <xsl:variable name="func_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/method[@name = $func_name]/@alias"/>

        <!-- if we found a class name alias, then we use that, else the original -->

        <!-- now for the actual function name, one of 4 things (in
             order of priority):
              1) NEW if it's a constructor
              2) DELETE if it's a destructor
              3) an alias if one is supplied in the user options xml file
              4) the actual function name
          -->
        <xsl:choose>
          <!-- constructors are called NEW -->
          <xsl:when test="@ctor='yes'">
        <xsl:value-of select="$class_alias"/>
        <xsl:if test="string-length($class_alias) = 0">
          <xsl:value-of select="$class_name"/>
        </xsl:if>

        <xsl:text>_</xsl:text>
            <xsl:text>NEW</xsl:text>
          </xsl:when>

          <!-- destructors are called DELETE -->
          <xsl:when test="@dtor='yes'">
        <xsl:value-of select="$class_alias"/>
        <xsl:if test="string-length($class_alias) = 0">
          <xsl:value-of select="$class_name"/>
        </xsl:if>

        <xsl:text>_</xsl:text>
            <xsl:text>DELETE</xsl:text>
          </xsl:when>

          <!-- if we found a method name alias, then we use that, else the original -->
          <xsl:otherwise>
            <xsl:value-of select="$func_alias"/>
            <xsl:if test="string-length($func_alias) = 0">
              <xsl:value-of select="$func_name"/>
            </xsl:if>
          </xsl:otherwise>
        </xsl:choose>

        <!-- if this is a class method, then we append a number to the end -->
        <xsl:value-of select="position()"/>    
        

        <!-- the unique class id -->
        <xsl:value-of select="$class_id"/>
        
        

      </xsl:if>

      <!-- if this is not a class method, then just print the name -->
      <xsl:if test="not(ancestor::structure)">
        <xsl:value-of select="$func_name"/>
      </xsl:if>
    
          <!-- if we're generating the final generic name, then append a
           "_F" if this function has a return type other than void or ref -->
 <!--     <xsl:if test="return/type/@kind != 'void' and return/type/@kind != 'ref'">
        <xsl:text>_F</xsl:text>
      </xsl:if>
-->

    </xsl:variable>

    <!-- runs the name through the dictionary renamer -->
<!--    <xsl:variable name="fixed_name">
      <xsl:call-template name="fixname">
        <xsl:with-param name="name" select="$final_name"/>
      </xsl:call-template>
    </xsl:variable>
-->
    <!-- so now we created our variable, so output it -->
    <xsl:variable name="upper_name">
      <xsl:value-of select="translate($final_name,'abcdefghijklmnopqrstuvwxyz','ABCDEFGHIJKLMNOPQRSTUVWXYZ')"/>
    </xsl:variable>

    <xsl:value-of select="$upper_name"/>

    <!-- and check to make sure that the length is legal -->
    <xsl:call-template name="check_identifier_length">
      <xsl:with-param name="identifier" select="$upper_name"/>
    </xsl:call-template>
    
  </xsl:template>



  <!--
    Generate the name of an argument.  If the argument has no name, e.g. '-',
    then we automatically generate a name like: arg1, arg2, arg3...
    -->
  <xsl:template name="gen_arg_name">
  
    <!-- since we need to check the resulting length of the generated name,
         we do all the output of this template into a variable, which we can
         just then check at the end -->
    <xsl:variable name="final_name">
  
    <xsl:if test="@name = '-'">
      <xsl:text>arg</xsl:text>
      <xsl:value-of select="position()"/>
    </xsl:if>    
    <xsl:if test="@name != '-'">
      <xsl:value-of select="@name"/>
    </xsl:if>    
  
    </xsl:variable>

    <!-- so now we created our variable, so output it -->
    <xsl:value-of select="$final_name"/>

    <xsl:variable name="fixed_name">
      <xsl:call-template name="fixname">
        <xsl:with-param name="name" select="$final_name"/>
      </xsl:call-template>
    </xsl:variable>

    <!-- and check to make sure that the length is legal -->
    <xsl:call-template name="check_identifier_length">
      <xsl:with-param name="identifier" select="$fixed_name"/>
    </xsl:call-template>
  
  </xsl:template>



  <!--
    Generate the wrapper functions.
    -->
  <xsl:template match="method" mode="decl_fc_c">
    <!-- every template that matches methods needs to first verify that
         it is valid and that a wrapper should be generated -->         
    <xsl:variable name="method_status">
      <xsl:call-template name="check_method"/>
    </xsl:variable>
    <xsl:if test="$method_status = 'pass'">

    <xsl:value-of select="$newline"/>
    <xsl:text>void </xsl:text>
    
    <xsl:call-template name="gen_func_name"/>     
    <xsl:text>_C( </xsl:text>
      
    <xsl:call-template name="decl_fc_c_class_args"/>
    <xsl:apply-templates select="arg|return" mode="decl_fc_c_args"/>
      
    <xsl:text> )</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:text>{</xsl:text>
    <xsl:value-of select="$newline"/>
    
    <!-- special handling is required for string arguments/returns -->
    <xsl:apply-templates select="arg" mode="handle_strings_c_pre"/>
    <xsl:apply-templates select="return" mode="handle_strings_c_pre"/>
    
    <xsl:choose>

      <xsl:when test="@kind='static'">
        <xsl:call-template name="decl_fc_c_static"/>
      </xsl:when>

      <!-- if it's a constructor, generate the special constructor/desctructor call -->
      <xsl:when test="@ctor='yes' or @dtor='yes'">
        <xsl:call-template name="decl_fc_c_ctor"/>
      </xsl:when>
     
       <!-- if it's a not a constructor, generate the simple call -->
      <xsl:when test="not(@ctor='yes') and not(@dtor='yes')">
        <xsl:call-template name="decl_fc_c_call"/>
      </xsl:when>

    </xsl:choose>
    
    <!-- special handling is required for string arguments/returns -->
    <xsl:apply-templates select="arg" mode="handle_strings_c_post"/>
    <xsl:apply-templates select="return" mode="handle_strings_c_post"/>
    
    <xsl:value-of select="$newline"/>
    <xsl:text>}</xsl:text>
    <xsl:value-of select="$dblnewline"/>
    
    </xsl:if>
  </xsl:template>



  <!--
    Generate the self object argument for any methods that are part of a
    class. It's always the first argument of any method that is part of a
    class.  
    
    Note that we need to prepend a class name if this class is defined
    inside another class (like if this class is nested within another
    class and the higher-level class has the definition).  This is not
    currently implemented.
    -->
  <xsl:template name="decl_fc_c_class_args">

    <xsl:if test="parent::structure">
      <xsl:if test="not(@kind='static')">
        <xsl:value-of select="parent::structure/@name"/>
        <xsl:text> **self</xsl:text>
      </xsl:if>
    </xsl:if>

  </xsl:template>



  <!--
    Generate the argument list for a wrapper function declaration, so
    it needs to have argument types and names, with extra length info
    for character arrays.
    -->
  <xsl:template match="arg|return" mode="decl_fc_c_args">

   <!--  we only add return parameters for functions that have return types,
         so don't do it if the function returns a 'ref' or a 'void', since
         these indicate that there is no real return value. 
         
         but, some arguments can be of type 'ref', so we need to allow them. -->
   <xsl:if test="not(type/@kind='ref') and not(type/@kind='void') or string-length(@name) &gt; 0">

    <!-- add a comma before if this is not the first argument
         (if this is in a class, then there's always a comma since
         there is always a class object as the first parameter,
         except for static methods, they don't have a self.) -->
    <xsl:if test="position() != 1 or (ancestor::structure and not(parent::method/@kind='static'))">
      <xsl:text>, </xsl:text>
    </xsl:if>
 
    <!-- generate the type of this argument in C -->
    <xsl:call-template name="xml_type_to_c_type"/>

    <!-- add a * before each variable name -->
    <xsl:text> *</xsl:text>
    
    <!-- if it's a pointer (but not a char*), then add another * -->
<!--    <xsl:if test="type/@kind='ptr' and not(descendant::type/@ikind='char')">
-->    <xsl:if test="type/@kind='ptr' and descendant::type/@kind='usertype'">
      <xsl:text>*</xsl:text>
    </xsl:if>    
    
    <!-- if it's a ref, then add a & -->
    <xsl:if test="type/@kind='ref' and descendant::indirect/type/@kind='ptr'">
      <xsl:text>&amp;</xsl:text>
    </xsl:if>    
    
    <!-- 
      add the variable name (note that return tags don't have names,
      so if there is no name attribute, then we must have matched to
      a return tag, in which case we just use the name "rtn"
      -->
    <xsl:call-template name="gen_arg_name"/>
    <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
    
    <!-- if it's a char array, then we need to add an extra length argument -->
    <xsl:if test="descendant::indirect/type/@ikind = 'char'">
      <xsl:text>, int *</xsl:text>
      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if>
      <xsl:text>_len</xsl:text>
    </xsl:if>

    </xsl:if>

   </xsl:template>



  <!--
    Convert a type in the xml file to its corresponding C type.
    -->
  <xsl:template name="xml_type_to_c_type">
    <!-- handle each type of kind attribute -->
    <xsl:choose>
    
      <!-- handle integer types -->
      <xsl:when test="type/@kind = 'int'">

        <!-- if it's an integer type, then look at the ikind attribute
             to see what kind of integer it is. -->
        <xsl:choose>
          <xsl:when test="type/@ikind = 'char'">
            <xsl:text>char</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'uchar'">
            <xsl:text>unsigned char</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'short'">
            <xsl:text>short</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'ushort'">
            <xsl:text>unsigned short</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'int'">
            <xsl:text>int</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'uint'">
            <xsl:text>unsigned int</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'long'">
            <xsl:text>long</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'ulong'">
            <xsl:text>unsigned long</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'longlong'">
            <xsl:text>long long</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'ulonglong'">
            <xsl:text>unsigned long long</xsl:text>
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>UNSUPPORTED</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>


      <!-- handle float types -->
      <xsl:when test="type/@kind = 'float'">

        <!-- if it's a float type, then look at the fkind attribute
             to see what kind of float it is. -->
        <xsl:choose>
          <xsl:when test="type/@fkind = 'float'">
            <xsl:text>float</xsl:text>
          </xsl:when>
          <xsl:when test="type/@fkind = 'dbl'">
            <xsl:text>double</xsl:text>
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>UNSUPPORTED</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      
      
      <!-- handle user types -->
      <xsl:when test="type/@kind = 'usertype'">
        <xsl:value-of select="type/@usertype"/>
      </xsl:when>


      <!-- handle the void type -->
      <xsl:when test="type/@kind = 'void'">
        <xsl:text>void</xsl:text>
      </xsl:when>


      <!-- handle the array type -->
      <xsl:when test="type/@kind = 'array'">
        <xsl:for-each select="type/array">
          <xsl:call-template name="xml_type_to_c_type"/>
        </xsl:for-each>
      </xsl:when>


      <!-- handle pointer types (by recursively calling this function) -->
      <xsl:when test="type/@kind = 'ptr'">
        <xsl:for-each select="type/indirect">
          <xsl:call-template name="xml_type_to_c_type"/>

          <xsl:if test="@depth='2'">
            <xsl:text>*</xsl:text>
          </xsl:if>

        </xsl:for-each>
      </xsl:when>
      

      <!-- handle the ref type.  i'm assuming that it'll always be a pointer? -->
      <xsl:when test="type/@kind = 'ref'">
        <xsl:for-each select="type/indirect">
          <xsl:call-template name="xml_type_to_c_type"/>
        </xsl:for-each>
      </xsl:when>


      <!-- warn about unsupported types -->
      <xsl:otherwise>
        <xsl:text>UNSUPPORTED</xsl:text>
      </xsl:otherwise>
    
    </xsl:choose>
  </xsl:template>



  <!--
    Generate the code to reallocate character arrays (this rule also matches
    return values).
    -->
  <xsl:template match="arg|return" mode="handle_strings_c_pre">
  
    <!-- for each argument that is a char* -->
    <xsl:if test="descendant::indirect/type/@ikind = 'char'">
  
      <!-- declare a new variable that is one char bigger -->
      <xsl:text>char *</xsl:text>
      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_local = new char[*</xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_len+1];</xsl:text>
      <xsl:value-of select="$newline"/>

      <!-- copy the fortran string into our c string -->
      <xsl:text>strcpy( </xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_local, </xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text> );</xsl:text>
      <xsl:value-of select="$newline"/>

      <!-- terminate the c string -->
      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_local[*</xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_len] = 0;</xsl:text>
      <xsl:value-of select="$newline"/>
      <xsl:value-of select="$newline"/>
 
    </xsl:if>
  
  </xsl:template>



  <!--
    Generate the call to the function that we're wrapping.  This should
    only be called for non-constructor, non-static methods.  Constructors
    and static methods are special cases and are handled by decl_fc_c_ctor()
    and decl_fc_c_static(). 
    -->
  <xsl:template name="decl_fc_c_call">
    
    <xsl:apply-templates select="return" mode="call_c_rtn_pre"/>
    
    <!-- if this method is in a class, then we need to use an
         object to reference it -->
    <xsl:if test="ancestor::structure">
      <xsl:text>(*self)-></xsl:text>
    </xsl:if>
    
    <xsl:value-of select="@name"/>    
    <xsl:text>( </xsl:text>    
    <xsl:apply-templates select="arg" mode="call_fc_c_call_args"/>
    <xsl:text>) </xsl:text>

    <xsl:apply-templates select="return" mode="call_c_rtn_post"/>

    <xsl:text>;</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:value-of select="$newline"/>
  
  </xsl:template>



  <!--
    Generate the call to the function that we're wrapping.  This should
    only be called for non-constructor, non-static methods.  Constructors
    and static methods are special cases and are handled by decl_fc_c_ctor()
    and decl_fc_c_static(). 
    
    this should only be called on methods of a class that are static.
    -->
  <xsl:template name="decl_fc_c_static">
    
    <xsl:apply-templates select="return" mode="call_c_rtn_pre"/>
    
    <xsl:value-of select="ancestor::structure/@name"/>    
    <xsl:text>::</xsl:text>    
    <xsl:value-of select="@name"/>    
    <xsl:text>( </xsl:text>    
    <xsl:apply-templates select="arg" mode="call_fc_c_call_args"/>
    <xsl:text>) </xsl:text>

    <xsl:apply-templates select="return" mode="call_c_rtn_post"/>

    <xsl:text>;</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:value-of select="$newline"/>
  
  </xsl:template>



  <!--
    Generate the call to the function that we're wrapping.  This should
    only be called for constructor methods.  Non-Constructors are
    handled by decl_fc_c_call()
    
    I'm not thrilled that this is in a separate template as normal functions
    and that an if-then decides which is called.
    -->
  <xsl:template name="decl_fc_c_ctor">
    
    <!-- if it's a constructor -->    
    <xsl:if test="@ctor='yes'">
      <xsl:text>(*self) = new </xsl:text>
      <xsl:value-of select="ancestor::structure/@name"/>    
    
      <xsl:text>( </xsl:text>    
      <xsl:apply-templates select="arg" mode="call_fc_c_call_args"/>
      <xsl:text>) </xsl:text>

      <xsl:text>;</xsl:text>
    </xsl:if>

    <!-- if it's a destructor -->    
    <xsl:if test="@dtor='yes'">
      <xsl:text>delete *self;</xsl:text>
    </xsl:if>

    <xsl:value-of select="$newline"/>
    <xsl:value-of select="$newline"/>
  
  </xsl:template>



  <!--
    Generate the arguments for the call the function we're wrapping.
    -->
  <xsl:template match="arg" mode="call_fc_c_call_args">

    <!-- add a comma before if this is not the first argument -->
    <xsl:if test="position() != 1">
      <xsl:text>, </xsl:text>
    </xsl:if>

    <!-- add a * before each non-char* variable name -->
<!--    <xsl:if test="not(descendant::type/@kind = 'ptr' and descendant::type/@ikind = 'char')">
 -->
<!--    <xsl:if test="not(type/@kind='ref') and descendant::type/@kind='usertype'">
  -->  <xsl:if test="not(descendant::type/indirect/@kind='pointer') or descendant::type/@kind='usertype'">
      <xsl:text>*</xsl:text>
    </xsl:if>
   
    <!-- if it's a ref, then add a & -->
    <xsl:if test="type/@kind='ref' and descendant::indirect/type/@kind='ptr'">
      <xsl:text>&amp;</xsl:text>
    </xsl:if>    
        
    <!-- write out the variable name -->
    <xsl:call-template name="gen_arg_name"/>

    <!-- change the destination array for character arrays -->
    <xsl:if test="descendant::indirect/type/@ikind = 'char'">
      <xsl:text>_local</xsl:text>
    </xsl:if>

  </xsl:template>



  <!--
    Add any extra return handling code before the C function (that
    we're wrapping) is called.  This template may need more work later
    on to support kinds of return types.

    Note that all return types aren't checked for here, but this
    doesn't need to do anything for some return types, which is why
    there isn't an "otherwise" catch all or error message for
    unhandled types.
    
    The order here is important, e.g. we want the character array case
    to catch character array pointers, we don't want them to fall
    through to the ptr case.
    -->
  <xsl:template match="return" mode ="call_c_rtn_pre">

    <xsl:choose>
   
      <!-- if the return is a character array -->
      <xsl:when test="descendant::indirect/type/@ikind = 'char'">  
        <xsl:text>strcpy( rtn_local, </xsl:text>
      </xsl:when>


      <!-- if the return is an int or float -->
      <xsl:when test="type/@kind = 'int' or type/@kind = 'float'">  
        <xsl:text>*rtn = </xsl:text>
      </xsl:when>

      <!-- if the return is a ptr -->
      <xsl:when test="type/@kind = 'ptr'">
      
      
        <xsl:if test="descendant::type/@kind='usertype'">
        <xsl:text>*</xsl:text>
      </xsl:if>
        

      
        <xsl:text>rtn = </xsl:text>
      </xsl:when>

    </xsl:choose>

  </xsl:template>



  <!--
    Add any extra return handling code after the C function (that
    we're wrapping) is called.
    -->
  <xsl:template match="return" mode ="call_c_rtn_post">

    <!-- if the return is a character array -->
    <xsl:if test="descendant::indirect/type/@ikind = 'char'">
      <xsl:text> )</xsl:text>
    </xsl:if>

  </xsl:template>



  <!--
    Generate the code to copy character arrays back into their original argument.
    -->
  <xsl:template match="arg|return" mode="handle_strings_c_post">

    <!-- for each argument that is a char* -->
    <xsl:if test="descendant::indirect/type/@ikind = 'char'">
    
      <!-- reset the original fortran array to all spaces -->
      <xsl:text>memset( </xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>, ' ', *</xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_len );</xsl:text>
      <xsl:value-of select="$newline"/>

      <!-- copy in our new c string into the original fortran string -->
      <xsl:text>strncpy( </xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>, </xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_local, strlen( </xsl:text>

      <xsl:call-template name="gen_arg_name"/>

      <xsl:if test="not(@name)"><xsl:text>rtn</xsl:text></xsl:if> 
      <xsl:text>_local));</xsl:text>
      <xsl:value-of select="$newline"/>
  
    </xsl:if>
  
  </xsl:template>



  <!--
    Generate the include statements (might be a problem here determining
    which ones are necessary, e.g. when the functions use other classes
    as arguments).
    -->
  <xsl:template name="generate_c_incs">
    <xsl:text>#include &lt;string.h&gt;</xsl:text>
    <xsl:value-of select="$newline"/>

    <xsl:text>#include "</xsl:text>
    <xsl:value-of select="@name"/>
    <xsl:text>"</xsl:text>
    <xsl:value-of select="$newline"/>

    <xsl:text>#include "header.h"</xsl:text>

    <xsl:value-of select="$newline"/>
    <xsl:value-of select="$newline"/>
  </xsl:template>



  <!--
    Print file header information.
    -->
  <xsl:template name="open_file">

    <xsl:text>//</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:text>// fortran calling c - this file is automatically generated</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:text>//</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:value-of select="$newline"/>

  </xsl:template>



  <!--
    Finish up.
    -->
  <xsl:template name="close_file">
    <xsl:value-of select="$newline"/>
  </xsl:template>



  <!--
    Create extern "C" declaration.
    -->
  <xsl:template name="extern_c_pre">

    <xsl:text>extern "C" {</xsl:text>
    <xsl:value-of select="$newline"/>

  </xsl:template>



  <!--
    Close extern "C" declaration.
    -->
  <xsl:template name="extern_c_post">

    <xsl:text>};</xsl:text>
    <xsl:value-of select="$newline"/>

  </xsl:template>

</xsl:stylesheet>
