<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">

<!--
  This file contains the templates specific to generating the fortran wrapper code.
  -->

  <xsl:import href="common.xsl"/>
  <xsl:output method="text"/>



 <!--
    Root level.  This gets things going.
    -->
  <xsl:template match="/">
    <xsl:apply-templates select="library"/>
  </xsl:template>



  <!--
    Library level (is there ever more than one library?).  This is the
    highest level, and calls templates that match methods.  Any template
    that matches method elements must first check to make sure that
    a wrapper should be generated for that method (note that this code is
    the same for each template that matches method elements).
    -->
  <xsl:template match="library">
    <xsl:call-template name="open_file"/>

    <xsl:text>module </xsl:text>
    <xsl:call-template name="gen_module_name"/>
    <xsl:value-of select="$dblnewline"/>

    <!-- first generate all the use statements so that any user defined type
         that is used as a method argument is "use"'d -->
    <xsl:apply-templates select="scope/descendant::method" mode="generate_use"/>
    
    <!-- generate the use statement that uses the fortran type associated
         with this class. --> 
    <xsl:apply-templates select="scope/structure" mode="generate_use_self"/>
    
    <!-- generate the interface aliases for all the wrapper functions -->
    <xsl:apply-templates select="scope/descendant::method" mode="generate_aliases"/>

    <xsl:text>contains</xsl:text>
    <xsl:value-of select="$newline"/>

    <!-- generate the actual wrapper functions -->
    <xsl:apply-templates select="scope/descendant::method" mode="decl_fc_f"/>

    <xsl:text>end module </xsl:text>
    <xsl:call-template name="gen_module_name"/>
    <xsl:value-of select="$newline"/>

    <xsl:call-template name="close_file"/>
  </xsl:template>



  <!--
    Generate the use statements.  This only creates use statements for
    usertypes that are an argument of a method, and also assumes that
    is should put an _type on the end of the struc type name.
    -->
  <xsl:template match="method" mode="generate_use">
    <!-- every template that matches methods needs to first verify that
         it is valid and that a wrapper should be generated -->         
    <xsl:variable name="method_status">
      <xsl:call-template name="check_method"/>
    </xsl:variable>
    <xsl:if test="$method_status = 'pass'">

    <!-- for each argument, see if it is a user-defined argument and
         if it is, generate a use statement for it. -->
    <xsl:for-each select="arg">
      <xsl:if test="type/@kind='usertype'">
        <xsl:text>use </xsl:text>
        <xsl:call-template name="gen_type_name_f">
          <xsl:with-param name="append_type" select="'yes'"/>
          <xsl:with-param name="type_name" select="type/@usertype"/>
        </xsl:call-template>
        <xsl:value-of select="$newline"/>
      </xsl:if>
    </xsl:for-each>

    </xsl:if>
  </xsl:template>



  <!--
    Since we declare the class type in a separate module definition, then
    we need to add a use statement to that all our module functions can
    use the type.
    -->
  <xsl:template match="structure" mode="generate_use_self">
    <xsl:text>use </xsl:text>

    <xsl:call-template name="gen_type_name_f">
      <xsl:with-param name="append_type" select="'yes'"/>
      <xsl:with-param name="type_name" select="@name"/>
    </xsl:call-template>

    <xsl:value-of select="$newline"/>
  </xsl:template>



  <!--
    Generate aliases for all functions.
  
    Generate all the aliases for functions that have multiple argument
    types (ie. void* arguments).
    -->
  <xsl:template match="method" mode="generate_aliases">
    <!-- every template that matches methods needs to first verify that
         it is valid and that a wrapper should be generated -->         
    <xsl:variable name="method_status">
      <xsl:call-template name="check_method"/>
    </xsl:variable>
    <xsl:if test="$method_status = 'pass'">

      <!-- for each method generate an alias for that method -->

      <!-- open the interface section -->
      <xsl:text>interface </xsl:text>
      <xsl:call-template name="gen_func_name">
        <xsl:with-param name="kind" select="'visible'"/>
      </xsl:call-template>
      <xsl:value-of select="$newline"/>
  
      <!-- call the template to generate the aliases -->
      <xsl:call-template name="generate_functions">
        <xsl:with-param name="mode" select="'alias'"/>
      </xsl:call-template>

      <!-- close the interface section -->
      <xsl:text>end interface </xsl:text>  
      <xsl:call-template name="gen_func_name">
        <xsl:with-param name="kind" select="'visible'"/>
      </xsl:call-template>
      <xsl:value-of select="$dblnewline"/>

    </xsl:if>
  </xsl:template>



  <!--
    Either generate the aliases for the functions, or generate the
    functions themselves, depending on the mode.
    -->
  <xsl:template name="generate_functions">
    <xsl:param name="mode"/>
    <xsl:param name="i" select="'1'"/>

    <!-- count how many void* arguments there are and save the value
      in a variable called voidcount -->
    <xsl:variable name="voidcount" select="count(arg/descendant::type[@kind='void'])"/>

    <!-- if the return type of this method is a void*, then we need
         to call the helper for each type we want to support.  this
         unfortunately means we need another parameter that will be
         passed all the way down. -->

    <!-- generate the next void* type parameter -->
    <xsl:variable name="voidreturntype">
      <!-- if it has a void* return type -->
      <xsl:if test="descendant::return/type/indirect/type/@kind='void'">
        <xsl:value-of select="substring( $datatypes, $i, 1 )"/>
      </xsl:if>
    </xsl:variable>    

    <!-- now call a template to generate the function for each type,
         passing in the total void* count -->
    <xsl:call-template name="generate_functions_helper">
      <xsl:with-param name="prefix" select="''"/>
      <xsl:with-param name="voidcount" select="$voidcount"/>
      <xsl:with-param name="i" select="'1'"/>
      <xsl:with-param name="mode" select="$mode"/>
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:call-template>

    <xsl:if test="descendant::return/type/indirect/type/@kind='void'">
      <xsl:if test="$i &lt; string-length( $datatypes )">
        <xsl:call-template name="generate_functions">
          <xsl:with-param name="i" select="$i + 1"/>
          <xsl:with-param name="mode" select="$mode"/>
        </xsl:call-template>
      </xsl:if>
    </xsl:if>

  </xsl:template>



  <!--
    This actually figures out what all the alias names should be.
    It is based on the number of types we want to support, and the
    number of void* parameters that there are in the method.  For
    example, if we want void* arguments to be passed both integers
    and reals, and the method has 2 void* parameters, then this
    template generates 4 aliases to cover all the possible combinations
    of datatypes for the two void* parameters.  This basically results
    in d^v aliases, where d is the number of datatypes, and v is
    the number of void* parameters in the method.
    
    Furthermore, XSL doesn't support loops, so the loop that enumerate
    all the permutations is implemented using two mutually recursive
    templates that call each other in order to fake the loop.
    
    Parameters:
       o prefix -       always starts empty, stores the result
       o voidcount -    how many void* arguments the method has 
       o i -            starts at 1, used to iterate through the
                        datatypes string
       o mode -         "alias" if the just the alias should be
                        generated, or "decl" for the whole function
       o voidreturntype -    some extra text that gets appended to the
                        final function name
    -->
  <xsl:template name="generate_functions_helper">
    <xsl:param name="prefix"/>
    <xsl:param name="voidcount"/>
    <xsl:param name="i"/>
    <xsl:param name="mode"/>
    <xsl:param name="voidreturntype"/>

    <!-- if the prefix string contains a finished permutation, then print
         it, otherwise enter the loop to finish the permutation. -->
    <xsl:choose>
      <xsl:when test="string-length($prefix) = $voidcount">
      
        <!-- when we're done generating the next name permutation: -->
      
        <!-- if we're generating aliases, then do it -->
        <xsl:if test="$mode = 'alias'">
          <xsl:text>module procedure </xsl:text>  
          <xsl:call-template name="gen_func_name">
            <xsl:with-param name="arraytype" select="$prefix"/>
            <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
          </xsl:call-template>
          <xsl:value-of select="$newline"/>
        </xsl:if>
        
        <!-- if we're generating full functions, then do it -->
        <xsl:if test="$mode = 'decl'">
        
          <!-- if this is a constructor -->
          <xsl:if test="@ctor='yes'">
            <xsl:call-template name="decl_fc_f_helper_ctor">
              <xsl:with-param name="arraytype" select="$prefix"/>
              <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
            </xsl:call-template>
          </xsl:if>

          <!-- if this is not a constructor -->
          <xsl:if test="not(@ctor='yes')">
            <xsl:call-template name="decl_fc_f_helper">
              <xsl:with-param name="arraytype" select="$prefix"/>
              <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
            </xsl:call-template>
          </xsl:if>
        </xsl:if>
        
      </xsl:when>
      <xsl:otherwise>

        <!-- if we're not done permuting the next function name, loop again -->

        <xsl:call-template name="generate_functions_loop">
          <xsl:with-param name="prefix" select="$prefix"/>
          <xsl:with-param name="voidcount" select="$voidcount"/>
          <xsl:with-param name="i" select="$i"/>
          <xsl:with-param name="mode" select="$mode"/>
          <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
 
  </xsl:template>



  <!--
    This is the helper template for the templates that need to
    generate all the function names and is used to implement
    a loop.  Same parameters as generate_functions_helper().
    -->
  <xsl:template name="generate_functions_loop">
    <xsl:param name="prefix"/>
    <xsl:param name="voidcount"/>
    <xsl:param name="i"/>
    <xsl:param name="mode"/>
    <xsl:param name="voidreturntype"/>

    <!-- if we haven't gone through all the datatypes, then get the
       next type letter and send a new prefix back to the
       generate_alias() template.  then make a recursive call to
       continue the for loop. -->
    <xsl:if test="$i &lt; string-length($datatypes) + 1">    

      <xsl:variable name="c" select="substring( $datatypes, $i, 1 )"/>

      <xsl:call-template name="generate_functions_helper">
        <xsl:with-param name="prefix" select="concat($prefix,$c)"/>
        <xsl:with-param name="voidcount" select="$voidcount"/>
        <xsl:with-param name="i" select="'1'"/>
        <xsl:with-param name="mode" select="$mode"/>
        <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
      </xsl:call-template>

      <xsl:call-template name="generate_functions_loop">
        <xsl:with-param name="prefix" select="$prefix"/>
        <xsl:with-param name="voidcount" select="$voidcount"/>
        <xsl:with-param name="i" select="$i+1"/>
        <xsl:with-param name="mode" select="$mode"/>
        <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
      </xsl:call-template>

    </xsl:if>
 
  </xsl:template>



  <!--
    Generate the wrapper functions.  This function is matched to every
    method in the xml file.  For each one, it must check to see if the
    method has any void* arguments.
    -->
  <xsl:template match="method" mode="decl_fc_f">
    <!-- every template that matches methods needs to first verify that
         it is valid and that a wrapper should be generated -->         
    <xsl:variable name="method_status">
      <xsl:call-template name="check_method"/>
    </xsl:variable>
    <xsl:if test="$method_status = 'pass'">

    <!-- call the template to generate the functions -->
    <xsl:call-template name="generate_functions">
      <xsl:with-param name="mode" select="'decl'"/>
    </xsl:call-template>

  </xsl:if>
  </xsl:template>



  <!--
    This generates the module name based on the current header name.  E.g:
    headers/MyHeader.h becomes MyHeaderMod  This template also removes any
    illegal characters for module names.
   -->
  <xsl:template name="gen_module_name">

    <xsl:variable name="modname">

      <!-- first grab just the filename from the path -->
      <xsl:variable name="filename">
        <xsl:call-template name="getfilename">
          <xsl:with-param name="path" select="@name"/>
        </xsl:call-template>
      </xsl:variable>

      <!-- if it contains a .h, then replace that with Mod -->
      <xsl:if test="contains( $filename, '.h' )">
        <xsl:value-of select="substring-before( $filename, '.h' )"/>
        <xsl:text>Mod</xsl:text>
      </xsl:if>
      
    </xsl:variable>
   
    <!-- remove any illegal characters -->    
    <xsl:value-of select="translate($modname,'._', '')"/>

  </xsl:template>



  <!--
    This template generates the name of a fortran wrapper function. It 
    starts with the method name and may add a class alias if necessary,
    as well as any extra datatype designations (like for void* functions).
    
    Aliasing is not supported for constructors/destructors function
    names (but does alias the class name).
    
    It also checks to make sure that the length of the name is not longer
    than the length specified in the user options xml file.
    
    The if the append_number parameter is 'yes' then the method number
    (its position) is appended to the function name (default is yes).
    
    If the prepend_class_names argument is set to 'yes', then this template
    will check to see if the user options say to prepend_class_names, and
    if they are both 'yes', then the class name is prepended.
    
    If a C function has a return type, and we're generating the final
    generic name then we always append "_F" for two reasons:
      1) the user knows that a function has a return type
         as an argument as the last parameter.
      2) this also avoid naming clashes where by making the
         return type a parameter might shadow another function
         that already has no return type and the same parameter.
         
         
    kind should either be "internal" or "visible"
         
  -->
  <xsl:template name="gen_func_name">
    <xsl:param name="arraytype"/>
    <xsl:param name="voidreturntype"/>
    <xsl:param name="kind" select="'internal'"/>

    <!-- since we need to check the resulting length of the generated name,
         we do all the output of this template into a variable, which we can
         just then check at the end -->
    <xsl:variable name="final_name">

      <!-- find out what the original function name is -->
      <xsl:variable name="func_name" select="@name"/>

      <!-- if this is a class method, then we need to look up the alias for
           the class from the user preferences file and then prefix the function
           name with it -->
      <xsl:if test="ancestor::structure">

        <!-- find out what the original class name is -->
        <xsl:variable name="class_name" select="ancestor::structure/@name"/>

        <!-- lookup the aliases for the class/function names -->
        <xsl:variable name="class_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/@alias"/>
        <xsl:variable name="class_id" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/@mangle"/>
        <xsl:variable name="func_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/method[@name = $func_name]/@alias"/>

        <!-- now for the actual function name, one of 4 things (in
             order of priority):
               1) NEW if it's a constructor
               2) DELETE if it's a destructor
               3) an alias if one is supplied in the user options xml file
               4) the actual function name
          -->
        <xsl:choose>
          <!-- constructors are called NEW -->
          <xsl:when test="@ctor='yes'">

            <!-- if we found a class name alias, then we use that, else the original -->
            <xsl:value-of select="$class_alias"/>
            <xsl:if test="string-length($class_alias) = 0">
              <xsl:value-of select="$class_name"/>
            </xsl:if>

            <xsl:text>_</xsl:text>
            <xsl:text>NEW</xsl:text>
          </xsl:when>

          <!-- destructors are called DELETE -->
          <xsl:when test="@dtor='yes'">

            <!-- if we found a class name alias, then we use that, else the original -->
            <xsl:value-of select="$class_alias"/>
            <xsl:if test="string-length($class_alias) = 0">
              <xsl:value-of select="$class_name"/>
            </xsl:if>

            <xsl:text>_</xsl:text>
            <xsl:text>DELETE</xsl:text>
          </xsl:when>

          <!-- if we found a method name alias, then we use that, else the original -->
          <xsl:otherwise>
            <xsl:value-of select="$func_alias"/>
            <xsl:if test="string-length($func_alias) = 0">
              <xsl:value-of select="$func_name"/>
            </xsl:if>
          </xsl:otherwise>

        </xsl:choose>
      
        <xsl:if test="$kind='internal'">
      
          <!-- if this is a class method, then we append a number to the end for the call -->
          <xsl:value-of select="position()"/>   

          <!-- the unique class id -->
          <xsl:value-of select="$class_id"/>
        </xsl:if>

      </xsl:if>

      <!-- if this is not a class method, then just print the name -->
      <xsl:if test="not(ancestor::structure)">
        <xsl:value-of select="$func_name"/>
      </xsl:if>
       
      <!-- add any datatype designations as necessary -->
      <xsl:if test="string-length($arraytype) &gt; 0">
        <xsl:text>_</xsl:text>
        <xsl:value-of select="$arraytype"/>
      </xsl:if>
    
      <!-- add any return datatype designations as necessary -->
      <xsl:if test="string-length($voidreturntype) &gt; 0">
        <xsl:text>_rtn</xsl:text>
        <xsl:value-of select="$voidreturntype"/>
      </xsl:if>


      <!-- if we're generating the final generic name, then append a
           "_F" if this function has a return type other than void or ref -->
      <xsl:if test="$kind='visible'">
        <xsl:if test="return/type/@kind != 'void' and return/type/@kind != 'ref'">
          <xsl:text>_F</xsl:text>
        </xsl:if>
      </xsl:if>

    </xsl:variable>

    <!-- so now we created our variable, so output it -->
    <xsl:value-of select="$final_name"/>

    <!-- and check to make sure that the length is legal -->
    <xsl:call-template name="check_identifier_length">
      <xsl:with-param name="identifier" select="$final_name"/>
    </xsl:call-template>

  </xsl:template>



  <!--
    Generate the name of an argument.  If the argument has no name, e.g. '-',
    then we automatically generate a name like: arg1, arg2, arg3...
    
    Also, if the argument begins with an underscore, then we just remove it.
    Note that this does not handled the case of 2 or more leading underscores, we'll
    fix that when we need to.  Also note that this may cause a naming conflict in
    some (rare) cases.
    -->
  <xsl:template name="gen_arg_name">

    <!-- since we need to check the resulting length of the generated name,
         we do all the output of this template into a variable, which we can
         just then check at the end -->
    <xsl:variable name="final_name">
  
      <!-- first see if the name is just a '-' -->
      <xsl:if test="@name = '-'">
        <xsl:text>arg</xsl:text>
        <xsl:value-of select="position()"/>
      </xsl:if>    
    
      <!-- if it actually has a name, then we have to check for leading '_' -->
      <xsl:if test="@name != '-'">
        <xsl:choose>
      
        <!-- if it starts with a '_', then just use everything after it -->
        <xsl:when test="starts-with(@name,'_')">
          <xsl:value-of select="substring-after(@name,'_')"/>      
        </xsl:when>
      
        <!-- if not, then it looks like we can use the actual name -->
        <xsl:otherwise>
          <xsl:value-of select="@name"/>
        </xsl:otherwise>
      
        </xsl:choose>
      </xsl:if>    
  
    </xsl:variable>

    <!-- so now we created our variable, so output it -->
    <xsl:value-of select="$final_name"/>

    <!-- and check to make sure that the length is legal -->
    <xsl:call-template name="check_identifier_length">
      <xsl:with-param name="identifier" select="$final_name"/>
    </xsl:call-template>
  
  </xsl:template>



  <!--
    This template helps out the generate_functions() template by
    providing the code to generate the function bodies if need be.
    Note that this should only be called for non-constructor functions,
    constructor code is generated by decl_fc_f_helper_ctor().
    -->
  <xsl:template name="decl_fc_f_helper">
    <xsl:param name="arraytype"/>
    <xsl:param name="voidreturntype"/>
    
    <xsl:value-of select="$newline"/>
    <xsl:text>subroutine </xsl:text>
    
    <!-- generate the function name -->
    <xsl:call-template name="gen_func_name">
      <xsl:with-param name="arraytype" select="$arraytype"/>
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:call-template>

    <xsl:text>( </xsl:text>
    
    <xsl:call-template name="decl_fc_f_class_args"/>
    <xsl:apply-templates select="arg" mode="decl_fc_f_args"/>
    <xsl:apply-templates select="return" mode="decl_fc_f_args_return"/>
   
    <xsl:text> )</xsl:text>
    <xsl:value-of select="$dblnewline"/>
 
    <!-- generate the use statements for any user datatype arguments in this method -->
    <xsl:call-template name="generate_use_for_method"/>
 
    <!-- declare the self object (and the use statement for it) for methods -->
    <xsl:call-template name="decl_fc_f_args_declaration_class"/>

    <!-- now we create the variable type declarations.  but, we have to pass
         in the datatype in case this is a void* argument. -->
    <xsl:apply-templates select="arg" mode="decl_fc_f_args_declaration">
      <xsl:with-param name="arraytype" select="$arraytype"/>
    </xsl:apply-templates>
    
    <xsl:apply-templates select="return" mode="decl_fc_f_args_declaration_return">
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:apply-templates>
    <xsl:value-of select="$newline"/>

    <xsl:call-template name="decl_fc_f_call"/>

    <xsl:text>end subroutine </xsl:text>

    <!-- generate the function name -->
    <xsl:call-template name="gen_func_name">
      <xsl:with-param name="arraytype" select="$arraytype"/>
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:call-template>

    <xsl:value-of select="$dblnewline"/>

  </xsl:template>



  <!--
    Generate all the use statements for each of the user datatypes used as
    arguments or return values in this method.
    
    Actually, is this really necessary? or are all these already included
    in the module's includes?
    -->
  <xsl:template name="generate_use_for_method">

    <xsl:for-each select="arg|return">
      <xsl:if test="descendant::type/@kind = 'usertype'">
        <xsl:text>use </xsl:text>

        <xsl:call-template name="gen_type_name_f">
          <xsl:with-param name="append_type" select="'yes'"/>
          <xsl:with-param name="type_name" select="descendant::type/@usertype"/>
        </xsl:call-template>

        <xsl:value-of select="$newline"/>
      </xsl:if>
    </xsl:for-each>

  </xsl:template>



  <!--
    Generate the self object argument for any methods that are part of a class.
    It's always the first argument of any method that is part of a class.
    
    Hmmm, this may need additional work also if the self type is really defined
    in another class (like if this class is nested within another and the type
    is actually defined in one of the more outer classes).
    -->
  <xsl:template name="decl_fc_f_class_args">

    <xsl:if test="not(@kind = 'static')">
      <xsl:if test="ancestor::structure">
        <xsl:text>self</xsl:text>
      </xsl:if>
    </xsl:if>

  </xsl:template>



  <!--
    This template helps out the generate_functions() template by
    providing the code to generate the function bodies if need be.
    Note that this should only be called for constructor functions,
    non-constructor code is generated by decl_fc_f_helper().
    
    I don't really like the fact this is a separate method from the
    above very similar one, so maybe combine them at some point.
    Having some problems with keeping the two templates with consistent
    code.
    -->
  <xsl:template name="decl_fc_f_helper_ctor">
    <xsl:param name="arraytype"/>
    <xsl:param name="voidreturntype"/>
   
    <xsl:value-of select="$newline"/>
    
    <!-- generate the return type declaration -->
    <xsl:text>type(</xsl:text>
    
    <xsl:call-template name="gen_type_name_f">
      <xsl:with-param name="type_name" select="ancestor::structure/@name"/>
    </xsl:call-template>

    <xsl:text>) </xsl:text>
    
    <xsl:text>function </xsl:text>

    <!-- generate the function name -->
    <xsl:call-template name="gen_func_name">
      <xsl:with-param name="arraytype" select="$arraytype"/>
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:call-template>

    <xsl:text>( </xsl:text>
    
    <xsl:call-template name="decl_fc_f_class_args"/>
    <xsl:apply-templates select="arg" mode="decl_fc_f_args"/>
    <xsl:apply-templates select="return" mode="decl_fc_f_args_return"/>
   
    <xsl:text> )</xsl:text>
    <xsl:value-of select="$dblnewline"/>
 
    <!-- generate the use statements for any user datatype arguments in this method -->
    <xsl:call-template name="generate_use_for_method"/>
 
    <!-- declare the self object (and the use statement for it) for methods -->
    <xsl:call-template name="decl_fc_f_args_declaration_class"/>

    <!-- now we create the variable type declarations.  but, we have to pass
         in the datatype in case this is a void* argument. -->
    <xsl:apply-templates select="arg" mode="decl_fc_f_args_declaration">
      <xsl:with-param name="arraytype" select="$arraytype"/>
    </xsl:apply-templates>

    <xsl:apply-templates select="return" mode="decl_fc_f_args_declaration_return">
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:apply-templates>
    <xsl:value-of select="$newline"/>

    <!-- generate the call -->
    <xsl:call-template name="decl_fc_f_call"/>

    <!-- set the return object -->
    <xsl:call-template name="gen_func_name">
      <xsl:with-param name="arraytype" select="$arraytype"/>
    </xsl:call-template>

    <xsl:text> = self</xsl:text>
    <xsl:value-of select="$dblnewline"/>

    <!-- end the function declaration -->
    <xsl:text>end function </xsl:text>

    <!-- generate the function name -->
    <xsl:call-template name="gen_func_name">
      <xsl:with-param name="arraytype" select="$arraytype"/>
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:call-template>

    <xsl:value-of select="$dblnewline"/>

  </xsl:template>



  <!--
    Generate the call to the C wrapper function.
    -->
  <xsl:template name="decl_fc_f_call">
    
    <xsl:text>call </xsl:text>
    <xsl:call-template name="gen_func_name">
      <xsl:with-param name="prepend_class_names" select="'yes'"/>
    </xsl:call-template>
    <xsl:text>_C( </xsl:text>
    
    <!-- if this is a class, then the first parameter is always a
         self object, unless it's a static method -->
    <xsl:if test="ancestor::structure">
      <xsl:if test="not(@kind = 'static')">
        <xsl:text>self</xsl:text>
      </xsl:if>
    </xsl:if>

    <xsl:apply-templates select="arg" mode="call_fc_f_call_args"/>
    <xsl:apply-templates select="return" mode="call_fc_f_call_args_return"/>
    <xsl:text>) </xsl:text>

    <xsl:value-of select="$dblnewline"/>

  </xsl:template>



  <!--
    Generate the list of arguments in a fortran subroutine call (no types).
    -->
  <xsl:template match="arg" mode="call_fc_f_call_args">

    <!-- add a comma if necessary -->
    <xsl:call-template name="add_comma"/>

    <!-- if it's a pointer -->
    <xsl:if test="type/@kind = 'ptr'">

      <xsl:call-template name="gen_arg_name"/>
            
      <!-- if it's a character array, then we need to pass the length too -->
      <xsl:if test="descendant::indirect/type/@ikind = 'char'">
        <xsl:text>, len(</xsl:text>
        <xsl:call-template name="gen_arg_name"/>
        <xsl:text>)</xsl:text>
      </xsl:if>
    </xsl:if>

    <!-- if it's a user type -->
    <xsl:if test="type/@kind = 'usertype'">
      <xsl:call-template name="gen_arg_name"/>
    </xsl:if>
    
    <!-- if it's a ref type -->
    <xsl:if test="type/@kind = 'ref'">
      <xsl:call-template name="gen_arg_name"/>
    </xsl:if>
    
    <!-- if it's an int -->
    <xsl:if test="type/@kind = 'int'">
      <xsl:call-template name="gen_arg_name"/>
    </xsl:if>

    <!-- if it's a real -->
    <xsl:if test="type/@kind = 'float'">
      <xsl:call-template name="gen_arg_name"/>
    </xsl:if>

  </xsl:template>



  <!--
    This generates a comma (like for in between arguments) if there should
    be one for the current return/arg element.
    -->
  <xsl:template name="add_comma">
    <xsl:choose>

      <!-- if the position is greater than 1, then always add a comma, cus
           there must be arguments before this one. -->
      <xsl:when test="position() != 1">
        <xsl:text>, </xsl:text>
      </xsl:when>

      <!-- if this is an arg/return that is in a non-static method, then
           we always add a comma since all non-static methods have the
           self pointer as their first argument. -->
      <xsl:when test="ancestor::structure and not(parent::method/@kind='static')">
        <xsl:text>, </xsl:text>
      </xsl:when>

      <!-- if this is called from a return context, and there's other
           arguments in this method, then we add a comma. -->
      <xsl:when test="not(@name) and count(parent::method/arg) &gt; 0">
        <xsl:text>, </xsl:text>
      </xsl:when>

    </xsl:choose>
  </xsl:template>



  <!--
    Add any return value arguments to the end of a fortran subroutine call.
    This template will need additional work in order to support other
    return types.
    -->
  <xsl:template match="return" mode ="call_fc_f_call_args_return">

    <!-- if the return is a character array -->
    <xsl:if test="descendant::indirect/type/@ikind = 'char'">
      <xsl:call-template name="add_comma"/>
      <xsl:text>rtn, len(rtn) </xsl:text>
    </xsl:if>

    <!-- if the return is an integer or float -->
    <xsl:if test="type/@kind = 'int' or type/@kind = 'float'">
      <xsl:call-template name="add_comma"/>
      <xsl:text>rtn </xsl:text>
    </xsl:if>

  </xsl:template>



  <!--
    Generate the argument type declarations.  It includes an arraytype in
    case this argument is a void*, cus otherwise we can't tell what the
    type should be.
    -->
  <xsl:template match="arg" mode="decl_fc_f_args_declaration">
    <xsl:param name="arraytype"/>
    <xsl:param name="voidreturntype"/>
    
    <!-- convert the xml type to a fortran type -->
    <xsl:call-template name="xml_type_to_f_type">
      <xsl:with-param name="arraytype" select="$arraytype"/>
      <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
    </xsl:call-template>  
    
    <xsl:text> :: </xsl:text>
    <!-- now add the argument name -->
    <xsl:call-template name="gen_arg_name"/>
    <xsl:value-of select="$newline"/>

  </xsl:template>



  <!--
    Generate the argument type declarations for the self object.  
    
    Hmmm... since the module should already be including this type,
    is this code necessary?
    -->
  <xsl:template name="decl_fc_f_args_declaration_class">
    <xsl:param name="arraytype"/>

    <!-- if this is in a class... -->
    <xsl:if test="ancestor::structure">

      <xsl:text>use </xsl:text>
      <xsl:call-template name="gen_type_name_f">
        <xsl:with-param name="type_name" select="ancestor::structure/@name"/>
        <xsl:with-param name="append_type" select="'yes'"/>
      </xsl:call-template>

      <xsl:value-of select="$newline"/>

      <xsl:if test="not(@kind = 'static')">
        <xsl:text>type(</xsl:text>

        <xsl:call-template name="gen_type_name_f">
          <xsl:with-param name="type_name" select="ancestor::structure/@name"/>
        </xsl:call-template>

        <xsl:text>) :: self</xsl:text>
      </xsl:if>
      
      <xsl:value-of select="$newline"/>
    
    </xsl:if>
  </xsl:template>



  <!--
    Generate the return argument type declaration.
    -->
  <xsl:template match="return" mode="decl_fc_f_args_declaration_return">
    <xsl:param name="voidreturntype"/>

    <!-- if the return type is not a reference (like for constructors)
         or void (indicating there is no return type) then we declare it. -->
    <xsl:if test="not(type/@kind = 'ref') and not(type/@kind = 'void')">
      <xsl:call-template name="xml_type_to_f_type">
        <xsl:with-param name="voidreturntype" select="$voidreturntype"/>
      </xsl:call-template>
      
      <xsl:text> :: rtn</xsl:text>
      <xsl:value-of select="$newline"/>
    </xsl:if>

  </xsl:template>



  <!--
    Convert a type in the xml file to its corresponding fortran type.
    These are not all correct right now.  What are the correct types?
    -->
  <xsl:template name="xml_type_to_f_type">
    <xsl:param name="arraytype"/>
    <xsl:param name="voidreturntype"/>

    <!-- handle each type of kind attribute -->
    <xsl:choose>
    
      <!-- handle integer types -->
      <xsl:when test="type/@kind = 'int'">

        <!-- if it's an integer type, then look at the ikind attribute
             to see what kind of integer it is. -->
        <xsl:choose>
          <xsl:when test="type/@ikind = 'char'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'uchar'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'short'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'ushort'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'int'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'uint'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'long'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'ulong'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'longlong'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:when test="type/@ikind = 'ulonglong'">
            <xsl:text>integer</xsl:text>
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>UNSUPPORTED</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>


      <!-- handle the ref type.  i'm assuming that it'll always be a pointer? -->
      <xsl:when test="type/@kind = 'ref'">
        <xsl:for-each select="type/indirect">
          <xsl:call-template name="xml_type_to_f_type"/>
        </xsl:for-each>
      </xsl:when>


      <!-- handle float types -->
      <xsl:when test="type/@kind = 'float'">

        <!-- if it's a float type, then look at the fkind attribute
             to see what kind of float it is. -->
        <xsl:choose>
          <xsl:when test="type/@fkind = 'float'">
            <xsl:text>real(kind=selected_real_kind(p=6,r=30))</xsl:text>
          </xsl:when>
          <xsl:when test="type/@fkind = 'dbl'">
            <xsl:text>real(kind=selected_real_kind(p=8,r=30))</xsl:text>
          </xsl:when>
          <xsl:otherwise>
            <xsl:text>UNSUPPORTED</xsl:text>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      
      
      <!-- handle user types -->
      <xsl:when test="type/@kind = 'usertype'">
        <xsl:text>type(</xsl:text>
        
        <xsl:call-template name="gen_type_name_f">
          <xsl:with-param name="type_name" select="type/@usertype"/>
        </xsl:call-template>
        
        <xsl:text>)</xsl:text>
      </xsl:when>


      <!-- handle the void type -->
      <xsl:when test="type/@kind = 'void'">
        <xsl:text>void</xsl:text>
      </xsl:when>


      <!-- handle pointer types -->
      <xsl:when test="type/@kind = 'ptr'">
        
        <xsl:choose>
        
          <!-- if it's a character array, then we declare it as such -->
          <xsl:when test="descendant::indirect/type/@ikind = 'char'">
            <xsl:text>character(*)</xsl:text>
          </xsl:when>
        
          <!-- if it's a user type, then we declare it as such 
               prolly want to combine this with the usertype case above -->
          <xsl:when test="descendant::indirect/type/@kind = 'usertype'">
            <xsl:text>type(</xsl:text>

            <xsl:call-template name="gen_type_name_f">
              <xsl:with-param name="type_name" select="descendant::indirect/type/@usertype"/>
            </xsl:call-template>
            
            <xsl:text>)</xsl:text>
          </xsl:when>
        
          <!-- if it's some other type of array, then we use dimension() -->
          <xsl:otherwise>
            <!-- if it has an ikind, then use that for the pointer type, but
                 if not, then use the kind attribute -->
            <xsl:choose>

              <!-- if it's a real type, then step in, and call again -->
              <xsl:when test="descendant::indirect/type/@ikind">
                <xsl:for-each select="type/indirect">
                  <xsl:call-template name="xml_type_to_f_type"/>
                </xsl:for-each>
              </xsl:when>
              
              <!-- if it's a real type, then step in, and call again -->
              <xsl:when test="descendant::indirect/type/@fkind">
                <xsl:for-each select="type/indirect">
                  <xsl:call-template name="xml_type_to_f_type"/>
                </xsl:for-each>
              </xsl:when>
              
              <!-- if the type is void*, then we look to the arraytype variable -->
              <xsl:when test="descendant::type/@kind = 'void'">
                <xsl:if test="$arraytype = 'i' or $voidreturntype = 'i'">
                  <xsl:text>integer</xsl:text>
                </xsl:if>
                <xsl:if test="$arraytype = 'r' or $voidreturntype = 'r'">
                  <xsl:text>real</xsl:text>
                </xsl:if>
              </xsl:when>

              <!-- it's an error if we can't find a type -->
              <xsl:otherwise>
                <xsl:text>UNSUPPORTED</xsl:text>
              </xsl:otherwise>
            </xsl:choose>
        
            <xsl:text>,dimension(:)</xsl:text>
          </xsl:otherwise>
        
        </xsl:choose>
        
      </xsl:when>

      <!-- warn about unsupported types -->
      <xsl:otherwise>
        <xsl:text>UNSUPPORTED</xsl:text>
      </xsl:otherwise>
    
    </xsl:choose>
  </xsl:template>



  <!--
    Generate the argument list for a wrapper function declaration, so
    it needs to have argument types and names, with extra length info
    for character arrays.
    -->
  <xsl:template match="arg" mode="decl_fc_f_args">

    <!-- add a comma if necessary -->
    <xsl:call-template name="add_comma"/>
    
    <xsl:call-template name="gen_arg_name"/>
     
  </xsl:template>



  <!--
    Add any extra arguments to the end for return values.  Only add them
    if the return type is not a 'ref' or a 'void' (refs are from
    constructors).
    -->
  <xsl:template match="return" mode="decl_fc_f_args_return">
    <xsl:if test="not(type/@kind = 'ref') and not(type/@kind = 'void')">

      <!-- add a comma if necessary -->
      <xsl:call-template name="add_comma"/>
  
      <xsl:text>rtn</xsl:text>
    </xsl:if>
  </xsl:template>



  <!--
    Print file header information.
    -->
  <xsl:template name="open_file">

    <xsl:text>!</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:text>! fortran calling c - this file is automatically generated</xsl:text>
    <xsl:value-of select="$newline"/>
    <xsl:text>!</xsl:text>
    <xsl:value-of select="$dblnewline"/>

  </xsl:template>



  <!--
    Finish up.
    -->
  <xsl:template name="close_file">
    <xsl:value-of select="$newline"/>
  </xsl:template>



</xsl:stylesheet>
